import unittest

from tests.constants import (
    clean_up,
    set_up_utils,
)


class TestSequence(unittest.TestCase):
    """
    These tests should always work with an online controller.
    """

    def setUp(self):
        self.hplc_controller = set_up_utils(242, offline=False, runs=False)

    def tearDown(self):
        clean_up(self.hplc_controller)


if __name__ == "__main__":
    unittest.main()



import os
import time
import warnings
from typing import Dict, List, Optional, Union

from result import Err, Ok, Result
from xsdata.formats.dataclass.parsers import XmlParser

from ....analysis.process_report import AgilentReport, ReportType
from ....control.controllers import CommunicationController
from ....generated import DadMethod, PumpMethod, SolventElement
from pychemstation.analysis.chromatogram import (
    TIME_FORMAT,
    AgilentChannelChromatogramData,
    AgilentHPLCChromatogram,
)
from ....utils.macro import Command
from ....utils.method_types import (
    HPLCMethodParams,
    MethodDetails,
    Param,
    PType,
    TimeTableEntry,
)
from ....utils.table_types import RegisterFlag, T, Table, TableOperation
from ..devices.injector import InjectorController
from .table import TableController


class MethodController(TableController):
    """
    Class containing method related logic
    """

    def __init__(
        self,
        controller: CommunicationController,
        src: str,
        data_dirs: List[str],
        table: Table,
        offline: bool,
        injector_controller: InjectorController,
    ):
        self.injector_controller = injector_controller
        self.data_files: List[str] = []
        super().__init__(
            controller=controller,
            src=src,
            data_dirs=data_dirs,
            table=table,
            offline=offline,
        )

    def check(self) -> str:
        time.sleep(2)
        self.send(Command.GET_METHOD_CMD)
        time.sleep(2)
        res = self.receive()
        if res.is_ok():
            return res.ok_value.string_response
        return "ERROR"

    def get_method_params(self) -> HPLCMethodParams:
        return HPLCMethodParams(
            organic_modifier=self.controller.get_num_val(
                cmd=TableOperation.GET_OBJ_HDR_VAL.value.format(
                    register=self.table_locator.register,
                    register_flag=RegisterFlag.SOLVENT_B_COMPOSITION,
                )
            ),
            flow=self.controller.get_num_val(
                cmd=TableOperation.GET_OBJ_HDR_VAL.value.format(
                    register=self.table_locator.register,
                    register_flag=RegisterFlag.FLOW,
                )
            ),
        )

    def get_row(self, row: int) -> TimeTableEntry:
        flow = None
        om = None

        try:
            flow = self.get_num(row, RegisterFlag.TIMETABLE_FLOW)
        except RuntimeError:
            pass
        try:
            om = self.get_num(row, RegisterFlag.TIMETABLE_SOLVENT_B_COMPOSITION)
        except RuntimeError:
            pass

        return TimeTableEntry(
            start_time=self.get_num(row, RegisterFlag.TIME),
            organic_modifer=om,
            flow=flow,
        )

    def get_timetable(self, rows: int):
        uncoalesced_timetable_rows = [self.get_row(r + 1) for r in range(rows)]
        timetable_rows = {}
        for row in uncoalesced_timetable_rows:
            time_key = str(row.start_time)
            if time_key not in timetable_rows.keys():
                timetable_rows[time_key] = TimeTableEntry(
                    start_time=row.start_time,
                    flow=row.flow,
                    organic_modifer=row.organic_modifer,
                )
            else:
                if row.flow:
                    timetable_rows[time_key].flow = row.flow
                if row.organic_modifer:
                    timetable_rows[time_key].organic_modifer = row.organic_modifer
        entries = list(timetable_rows.values())
        entries.sort(key=lambda e: e.start_time)
        return entries

    def load(self) -> MethodDetails:
        rows = self.get_num_rows()
        if rows.is_ok():
            method_name = self.get_method_name()
            timetable_rows = self.get_timetable(int(rows.ok_value.num_response))
            params = self.get_method_params()
            stop_time = self.get_stop_time()
            post_time = self.get_post_time()
            self.table_state = MethodDetails(
                name=method_name,
                timetable=timetable_rows,
                stop_time=stop_time,
                post_time=post_time,
                params=params,
            )
            return self.table_state
        else:
            raise RuntimeError(rows.err_value)

    def get_method_name(self):
        self.send(Command.GET_METHOD_CMD)
        res = self.receive()
        method_name = res.ok_value.string_response
        return method_name

    def get_post_time(self) -> Union[int, float]:
        return self.controller.get_num_val(
            cmd=TableOperation.GET_OBJ_HDR_VAL.value.format(
                register=self.table_locator.register,
                register_flag=RegisterFlag.POST_TIME,
            )
        )

    def get_stop_time(self) -> Union[int, float]:
        return self.controller.get_num_val(
            cmd=TableOperation.GET_OBJ_HDR_VAL.value.format(
                register=self.table_locator.register,
                register_flag=RegisterFlag.MAX_TIME,
            )
        )

    def get_total_runtime(self) -> Union[int, float]:
        """Returns total method runtime in minutes."""
        return self.get_post_time() + self.get_stop_time()

    def current_method(self, method_name: str):
        """
        Checks if a given method is already loaded into Chemstation. Method name does not need the ".M" extension.

        :param method_name: a Chemstation method
        :return: True if method is already loaded
        """
        self.send(Command.GET_METHOD_CMD)
        parsed_response = self.receive()
        return method_name in parsed_response

    def switch(self, method_name: str, alt_method_dir: Optional[str] = None):
        """
        Allows the user to switch between pre-programmed methods. No need to append '.M'
        to the end of the method name. For example. for the method named 'General-Poroshell.M',
        only 'General-Poroshell' is needed.

        :param method_name: any available method in Chemstation method directory
        :param alt_method_dir: directory where the method resides
        :raise IndexError: Response did not have expected format. Try again.
        :raise AssertionError: The desired method is not selected. Try again.
        """
        method_dir = self.src if not alt_method_dir else alt_method_dir
        self.send(
            Command.SWITCH_METHOD_CMD_SPECIFIC.value.format(
                method_dir=method_dir, method_name=method_name
            )
        )

        time.sleep(2)
        self.send(Command.GET_METHOD_CMD)
        time.sleep(2)
        res = self.receive()
        if res.is_ok():
            parsed_response = res.ok_value.string_response
            assert parsed_response == f"{method_name}.M", "Switching Methods failed."
        self.table_state = None

    def load_from_disk(self, method_name: str) -> MethodDetails:
        """
        Retrieve method details of an existing method. Don't need to append ".M" to the end. This assumes the
        organic modifier is in Channel B and that Channel A contains the aq layer. Additionally, assumes
         only two solvents are being used.

        :param method_name: name of method to load details of
        :raises FileNotFoundError: Method does not exist
        :return: method details
        """
        warnings.warn("This method is not actively maintained.")
        method_folder = f"{method_name}.M"
        method_path = os.path.join(
            self.src, method_folder, "AgilentPumpDriver1.RapidControl.MethodXML.xml"
        )
        dad_path = os.path.join(
            self.src,
            method_folder,
            "Agilent1200erDadDriver1.RapidControl.MethodXML.xml",
        )

        if os.path.exists(os.path.join(self.src, f"{method_name}.M")):
            parser = XmlParser()
            method = parser.parse(method_path, PumpMethod)
            dad = parser.parse(dad_path, DadMethod)

            organic_modifier: Optional[SolventElement] = None

            if len(method.solvent_composition.solvent_element) == 2:
                for solvent in method.solvent_composition.solvent_element:
                    if solvent.channel == "Channel_B":
                        organic_modifier = solvent

            self.table_state = MethodDetails(
                name=method_name,
                params=HPLCMethodParams(
                    organic_modifier=organic_modifier.percentage, flow=method.flow
                ),
                stop_time=method.stop_time.stop_time_value,
                post_time=method.post_time.post_time_value,
                timetable=[
                    TimeTableEntry(
                        start_time=tte.time,
                        organic_modifer=tte.percent_b,
                        flow=method.flow,
                    )
                    for tte in method.timetable.timetable_entry
                ],
                dad_wavelengthes=dad.signals.signal,
            )
            return self.table_state
        else:
            raise FileNotFoundError

    def edit(self, updated_method: MethodDetails, save: bool):
        """Updated the currently loaded method in ChemStation with provided values.

        :param updated_method: the method with updated values, to be sent to Chemstation to modify the currently loaded method.
        :param save: if false only modifies the method, otherwise saves to disk
        """
        self.table_state = updated_method
        # Method settings required for all runs
        self.update_method_params(
            new_flow=updated_method.params.flow,
            new_initial_om=updated_method.params.organic_modifier,
            new_stop_time=updated_method.stop_time,
            new_post_time=updated_method.post_time,
        )
        self.edit_method_timetable(updated_method.timetable)

        if save:
            self.send(
                Command.SAVE_METHOD_CMD.value.format(
                    commit_msg=f"saved method at {str(time.time())}"
                )
            )

    def edit_initial_om(self, new_om: Union[int, float]):
        initial_organic_modifier: Param = Param(
            val=new_om,
            chemstation_key=RegisterFlag.SOLVENT_B_COMPOSITION,
            ptype=PType.NUM,
        )
        self._update_param(initial_organic_modifier)

    def edit_flow(self, new_flow: Union[int, float]):
        flow: Param = Param(
            val=new_flow, chemstation_key=RegisterFlag.FLOW, ptype=PType.NUM
        )
        self._update_param(flow)

    def edit_stop_time(self, new_stop_time: Union[int, float]):
        stop_time: Param = Param(
            val=new_stop_time,
            chemstation_key=RegisterFlag.MAX_TIME,
            ptype=PType.NUM,
        )
        self._update_param(
            Param(
                val="Set", chemstation_key=RegisterFlag.STOPTIME_MODE, ptype=PType.STR
            )
        )
        self._update_param(stop_time)

    def edit_post_time(self, new_post_time: Union[int, float]):
        post_time: Param = Param(
            val=new_post_time,
            chemstation_key=RegisterFlag.POST_TIME,
            ptype=PType.NUM,
        )
        self._update_param(
            Param(val="Set", chemstation_key=RegisterFlag.POSTIME_MODE, ptype=PType.STR)
        )
        self._update_param(post_time)

    def update_method_params(
        self,
        new_flow: Union[int, float],
        new_initial_om: Union[int, float],
        new_stop_time: Union[int, float],
        new_post_time: Union[int, float],
    ):
        self.delete_table()
        self.edit_initial_om(new_initial_om)
        self.edit_flow(new_flow)
        if self.table_state.stop_time:
            self.edit_stop_time(new_stop_time)
        else:
            self._update_param(
                Param(
                    val="Off",
                    chemstation_key=RegisterFlag.STOPTIME_MODE,
                    ptype=PType.STR,
                )
            )
        if self.table_state.post_time:
            self.edit_post_time(new_post_time)
        else:
            self._update_param(
                Param(
                    val="Off",
                    chemstation_key=RegisterFlag.POSTIME_MODE,
                    ptype=PType.STR,
                )
            )

    def _update_param(self, method_param: Param):
        """Change a method parameter, changes what is visibly seen in Chemstation GUI.
         (changes the first row in the timetable)

        :param method_param: a parameter to update for currently loaded method.
        """
        register = self.table_locator.register
        setting_command = (
            TableOperation.UPDATE_OBJ_HDR_VAL
            if method_param.ptype == PType.NUM
            else TableOperation.UPDATE_OBJ_HDR_TEXT
        )
        if isinstance(method_param.chemstation_key, list):
            for register_flag in method_param.chemstation_key:
                self.send(
                    setting_command.value.format(
                        register=register,
                        register_flag=register_flag,
                        val=method_param.val,
                    )
                )
        else:
            self.send(
                setting_command.value.format(
                    register=register,
                    register_flag=method_param.chemstation_key,
                    val=method_param.val,
                )
            )
        time.sleep(2)
        self.download()

    def download(self):
        self.send("Sleep 1")
        self.sleepy_send("DownloadRCMethod PMP1")
        self.send("Sleep 1")

    def _edit_row(self, row: TimeTableEntry, first_row: bool = False):
        if first_row:
            if row.organic_modifer:
                self.add_row()
                self.add_new_col_text(
                    col_name=RegisterFlag.FUNCTION,
                    val=RegisterFlag.SOLVENT_COMPOSITION.value,
                )
                self.add_new_col_num(col_name=RegisterFlag.TIME, val=row.start_time)
                self.add_new_col_num(
                    col_name=RegisterFlag.TIMETABLE_SOLVENT_B_COMPOSITION,
                    val=row.organic_modifer,
                )
            if row.flow:
                self.add_row()
                self.get_num_rows()
                self._edit_row_text(
                    col_name=RegisterFlag.FUNCTION, val=RegisterFlag.FLOW.value
                )
                self.add_new_col_num(col_name=RegisterFlag.TIMETABLE_FLOW, val=row.flow)
                self._edit_row_num(col_name=RegisterFlag.TIMETABLE_FLOW, val=row.flow)
            self.download()
        else:
            if row.organic_modifer:
                self.add_row()
                self.get_num_rows()
                self._edit_row_text(
                    col_name=RegisterFlag.FUNCTION,
                    val=RegisterFlag.SOLVENT_COMPOSITION.value,
                )
                self._edit_row_num(col_name=RegisterFlag.TIME, val=row.start_time)
                self._edit_row_num(
                    col_name=RegisterFlag.TIMETABLE_SOLVENT_B_COMPOSITION,
                    val=row.organic_modifer,
                )
                self.download()
            if row.flow:
                self.add_row()
                self.get_num_rows()
                self._edit_row_text(
                    col_name=RegisterFlag.FUNCTION, val=RegisterFlag.FLOW.value
                )
                self._edit_row_num(col_name=RegisterFlag.TIMETABLE_FLOW, val=row.flow)
                self._edit_row_num(col_name=RegisterFlag.TIME, val=row.start_time)
                self.download()

    def edit_method_timetable(self, timetable_rows: List[TimeTableEntry]):
        self.get_num_rows()
        self.delete_table()
        res = self.get_num_rows()
        while not res.is_err():
            self.delete_table()
            res = self.get_num_rows()

        self.new_table()
        self.get_num_rows()

        for i, row in enumerate(timetable_rows):
            self._edit_row(row=row, first_row=i == 0)

    def stop(self):
        """
        Stops the method run. A dialog window will pop up and manual intervention may be required.
        """
        self.send(Command.STOP_METHOD_CMD)

    def run(
        self,
        experiment_name: str,
        add_timestamp: bool = True,
        stall_while_running: bool = True,
    ):
        """
        :param experiment_name: Name of the experiment
        :param stall_while_running: whether to stall or immediately return
        :param add_timestamp: if should add timestamp to experiment name
        """

        folder_name = ""
        hplc_is_running = False
        tries = 0
        while tries < 10 and not hplc_is_running:
            timestamp = time.strftime(TIME_FORMAT)
            self.send(
                Command.RUN_METHOD_CMD.value.format(
                    data_dir=self.data_dirs[0],
                    experiment_name=f"{experiment_name}_{timestamp}"
                    if add_timestamp
                    else experiment_name,
                )
            )
            folder_name = (
                f"{experiment_name}_{timestamp}.D"
                if add_timestamp
                else f"{experiment_name}.D"
            )
            hplc_is_running = self.check_hplc_is_running()
            tries += 1

        if not hplc_is_running:
            raise RuntimeError("Method failed to start.")

        self.data_files.append(os.path.join(self.data_dirs[0], folder_name))
        self.timeout = (self.get_total_runtime()) * 60

        if stall_while_running:
            run_completed = self.check_hplc_done_running()
            if run_completed.is_ok():
                self.data_files[-1] = run_completed.ok_value
            else:
                raise RuntimeError("Run error has occurred.")
        else:
            folder = self.fuzzy_match_most_recent_folder(self.data_files[-1])
            while folder.is_err():
                folder = self.fuzzy_match_most_recent_folder(self.data_files[-1])
            if folder.is_ok():
                self.data_files[-1] = folder.ok_value
            else:
                warning = f"Data folder {self.data_files[-1]} may not exist, returning and will check again after run is done."
                warnings.warn(warning)

    def fuzzy_match_most_recent_folder(self, most_recent_folder: T) -> Result[T, str]:
        if os.path.exists(most_recent_folder):
            return Ok(most_recent_folder)
        return Err("Folder not found!")

    def get_data(
        self, custom_path: Optional[str] = None
    ) -> AgilentChannelChromatogramData:
        custom_path = custom_path if custom_path else self.data_files[-1]
        self.get_spectrum_at_channels(custom_path)
        return AgilentChannelChromatogramData(**self.spectra)

    def get_data_uv(
        self, custom_path: Optional[str] = None
    ) -> Dict[str, AgilentHPLCChromatogram]:
        custom_path = custom_path if custom_path else self.data_files[-1]
        self.get_uv_spectrum(custom_path)
        return self.uv

    def get_report(
        self,
        custom_path: Optional[str] = None,
        report_type: ReportType = ReportType.TXT,
    ) -> List[AgilentReport]:
        custom_path = self.data_files[-1] if not custom_path else custom_path
        metd_report = self.get_report_details(custom_path, report_type)
        chrom_data: List[AgilentHPLCChromatogram] = list(
            self.get_data(custom_path).__dict__.values()
        )
        for i, signal in enumerate(metd_report.signals):
            possible_data = chrom_data[i]
            if len(possible_data.x) > 0:
                signal.data = possible_data
        return [metd_report]



"""
.. include:: README.md
"""

from .comm import CommunicationController
from .tables.method import MethodController
from .tables.sequence import SequenceController

__all__ = ["CommunicationController", "MethodController", "SequenceController"]



import os
import time
import warnings
from typing import Dict, List, Optional

from result import Err, Ok, Result
from typing_extensions import override

from ....analysis.process_report import AgilentReport, ReportType
from ....control.controllers.comm import CommunicationController
from pychemstation.analysis.chromatogram import (
    SEQUENCE_TIME_FORMAT,
    AgilentChannelChromatogramData,
    AgilentHPLCChromatogram,
)
from ....utils.macro import Command
from ....utils.sequence_types import (
    InjectionSource,
    SampleType,
    SequenceDataFiles,
    SequenceEntry,
    SequenceTable,
)
from ....utils.table_types import RegisterFlag, Table
from ....utils.tray_types import FiftyFourVialPlate, TenVialColumn
from .. import MethodController
from .table import TableController


class SequenceController(TableController):
    """
    Class containing sequence related logic
    """

    def __init__(
        self,
        controller: CommunicationController,
        method_controller: MethodController,
        src: str,
        data_dirs: List[str],
        table: Table,
        offline: bool,
    ):
        self.method_controller = method_controller
        self.data_files: List[SequenceDataFiles] = []
        super().__init__(
            controller=controller,
            src=src,
            data_dirs=data_dirs,
            table=table,
            offline=offline,
        )

    def load(self) -> SequenceTable:
        rows = self.get_num_rows()
        self.send(Command.GET_SEQUENCE_CMD)
        seq_name = self.receive()

        if rows.is_ok() and seq_name.is_ok():
            self.table_state = SequenceTable(
                name=seq_name.ok_value.string_response.partition(".S")[0],
                rows=[
                    self.get_row(r + 1) for r in range(int(rows.ok_value.num_response))
                ],
            )
            return self.table_state
        raise RuntimeError(rows.err_value)

    def get_row(self, row: int) -> SequenceEntry:
        sample_name = self.get_text(row, RegisterFlag.NAME)
        vial_location = int(self.get_num(row, RegisterFlag.VIAL_LOCATION))
        method = self.get_text(row, RegisterFlag.METHOD)
        num_inj = int(self.get_num(row, RegisterFlag.NUM_INJ))
        inj_vol = float(self.get_text(row, RegisterFlag.INJ_VOL))
        inj_source = InjectionSource(self.get_text(row, RegisterFlag.INJ_SOR))
        sample_type = SampleType(self.get_num(row, RegisterFlag.SAMPLE_TYPE))
        vial_enum = (
            TenVialColumn(vial_location)
            if vial_location <= 10
            else FiftyFourVialPlate.from_int(num=vial_location)
        )
        return SequenceEntry(
            sample_name=sample_name,
            vial_location=vial_enum,
            method=None if len(method) == 0 else method,
            num_inj=num_inj,
            inj_vol=inj_vol,
            inj_source=inj_source,
            sample_type=sample_type,
        )

    def check(self) -> str:
        time.sleep(2)
        self.send(Command.GET_SEQUENCE_CMD)
        time.sleep(2)
        res = self.receive()
        if res.is_ok():
            return res.ok_value.string_response
        return "ERROR"

    def switch(self, seq_name: str):
        """
        Switch to the specified sequence. The sequence name does not need the '.S' extension.

        :param seq_name: The name of the sequence file
        """
        self.send(f'_SeqFile$ = "{seq_name}.S"')
        self.send(f'_SeqPath$ = "{self.src}"')
        self.send(Command.SWITCH_SEQUENCE_CMD)
        time.sleep(2)
        self.send(Command.GET_SEQUENCE_CMD)
        time.sleep(2)
        parsed_response = self.receive().ok_value.string_response

        assert parsed_response == f"{seq_name}.S", "Switching sequence failed."
        self.table_state = None

    def edit(self, sequence_table: SequenceTable):
        """
        Updates the currently loaded sequence table with the provided table. This method will delete the existing sequence table and remake it.
        If you would only like to edit a single row of a sequence table, use `edit_sequence_table_row` instead.

        :param sequence_table:
        """
        self.table_state = sequence_table
        rows = self.get_num_rows()
        if rows.is_ok():
            existing_row_num = rows.ok_value.num_response
            wanted_row_num = len(sequence_table.rows)
            while existing_row_num != wanted_row_num:
                if wanted_row_num > existing_row_num:
                    self.add_row()
                elif wanted_row_num < existing_row_num:
                    self.delete_row(int(existing_row_num))
                self.send(Command.SAVE_SEQUENCE_CMD)
                existing_row_num = self.get_num_rows().ok_value.num_response
            self.send(Command.SWITCH_SEQUENCE_CMD)

            for i, row in enumerate(sequence_table.rows):
                self._edit_row(row=row, row_num=i + 1)
                self.sleep(1)
            self.send(Command.SAVE_SEQUENCE_CMD)
            self.send(Command.SWITCH_SEQUENCE_CMD)

    def _edit_row(self, row: SequenceEntry, row_num: int):
        """
        Edits a row in the sequence table. If a row does NOT exist, a new one will be created.

        :param row: sequence row entry with updated information
        :param row_num: the row to edit, based on 1-based indexing
        """
        num_rows = self.get_num_rows()
        if num_rows.is_ok():
            while num_rows.ok_value.num_response < row_num:
                self.add_row()
                self.send(Command.SAVE_SEQUENCE_CMD)
                num_rows = self.get_num_rows()
        if row.vial_location:
            loc = row.vial_location
            if isinstance(loc, TenVialColumn):
                loc = row.vial_location.value
            elif isinstance(loc, FiftyFourVialPlate):
                loc = row.vial_location.value()
            self._edit_row_num(
                row=row_num, col_name=RegisterFlag.VIAL_LOCATION, val=loc
            )
        if row.method:
            method_dir = self.method_controller.src
            possible_path = os.path.join(method_dir, row.method) + ".M\\"
            method = row.method
            if os.path.exists(possible_path):
                method = os.path.join(method_dir, row.method)
            self._edit_row_text(row=row_num, col_name=RegisterFlag.METHOD, val=method)
        if row.num_inj:
            self._edit_row_num(
                row=row_num, col_name=RegisterFlag.NUM_INJ, val=row.num_inj
            )
        if row.inj_vol:
            self._edit_row_text(
                row=row_num, col_name=RegisterFlag.INJ_VOL, val=row.inj_vol
            )
        if row.inj_source:
            self._edit_row_text(
                row=row_num, col_name=RegisterFlag.INJ_SOR, val=row.inj_source.value
            )
        if row.sample_name:
            self._edit_row_text(
                row=row_num, col_name=RegisterFlag.NAME, val=row.sample_name
            )
            if row.data_file:
                self._edit_row_text(
                    row=row_num, col_name=RegisterFlag.DATA_FILE, val=row.data_file
                )
            else:
                self._edit_row_text(
                    row=row_num, col_name=RegisterFlag.DATA_FILE, val=row.sample_name
                )
        if row.sample_type:
            self._edit_row_num(
                row=row_num,
                col_name=RegisterFlag.SAMPLE_TYPE,
                val=row.sample_type.value,
            )

        self.send(Command.SAVE_SEQUENCE_CMD)

    def run(self, stall_while_running: bool = True):
        """
        Starts the currently loaded sequence, storing data
        under the <data_dir>/<sequence table name> folder.
        Device must be ready.
        """
        self.controller.send(Command.SAVE_METHOD_CMD)
        self.controller.send(Command.SAVE_SEQUENCE_CMD)

        if not self.table_state:
            self.table_state = self.load()

        total_runtime = 0
        for entry in self.table_state.rows:
            curr_method_runtime = self.method_controller.get_total_runtime()
            loaded_method = self.method_controller.get_method_name().removesuffix(".M")
            method_path = entry.method.split(sep="\\")
            method_name = method_path[-1]
            if loaded_method != method_name:
                method_dir = (
                    "\\".join(method_path[:-1]) + "\\" if len(method_path) > 1 else None
                )
                self.method_controller.switch(
                    method_name=method_name, alt_method_dir=method_dir
                )
                curr_method_runtime = self.method_controller.get_total_runtime()
            total_runtime += curr_method_runtime

        timestamp = time.strftime(SEQUENCE_TIME_FORMAT)
        self.send(Command.RUN_SEQUENCE_CMD.value)
        self.timeout = total_runtime * 60

        if self.check_hplc_is_running():
            folder_name = f"{self.table_state.name} {timestamp}"
            data_file = SequenceDataFiles(
                dir=folder_name, sequence_name=self.table_state.name
            )
            self.data_files.append(data_file)

            if stall_while_running:
                run_completed = self.check_hplc_done_running()
                if run_completed.is_ok():
                    self.data_files[-1] = run_completed.ok_value
                else:
                    warnings.warn("Run may have not completed.")
        else:
            raise RuntimeError("Sequence run did not start.")

    @override
    def fuzzy_match_most_recent_folder(
        self, most_recent_folder: SequenceDataFiles
    ) -> Result[SequenceDataFiles, str]:
        if os.path.isdir(most_recent_folder.dir):
            subdirs = [x[0] for x in os.walk(most_recent_folder.dir)]
            potential_folders = sorted(
                list(filter(lambda d: most_recent_folder.dir in d, subdirs))
            )
            most_recent_folder.child_dirs = [
                f
                for f in potential_folders
                if most_recent_folder.dir in f and ".M" not in f and ".D" in f
            ]
            return Ok(most_recent_folder)

        try:
            potential_folders = []
            for d in self.data_dirs:
                subdirs = [x[0] for x in os.walk(d)]
                potential_folders = sorted(
                    list(filter(lambda d: most_recent_folder.dir in d, subdirs))
                )
                if len(potential_folders) > 0:
                    break
            assert len(potential_folders) > 0
            parent_dirs = []
            for folder in potential_folders:
                path = os.path.normpath(folder)
                split_folder = path.split(os.sep)
                if most_recent_folder.dir in split_folder[-1]:
                    parent_dirs.append(folder)
            parent_dir = sorted(parent_dirs, reverse=True)[0]

            potential_folders = []
            for d in self.data_dirs:
                subdirs = [x[0] for x in os.walk(d)]
                potential_folders = sorted(
                    list(filter(lambda d: parent_dir in d, subdirs))
                )
                if len(potential_folders) > 0:
                    break
            assert len(potential_folders) > 0
            most_recent_folder.child_dirs = [
                f
                for f in potential_folders
                if parent_dir in f and ".M" not in f and ".D" in f
            ]
            return Ok(most_recent_folder)
        except Exception as e:
            error = f"Failed to get sequence folder: {e}"
            return Err(error)

    def get_data_uv(
        self, custom_path: Optional[str] = None
    ) -> List[Dict[int, AgilentHPLCChromatogram]]:
        custom_path = (
            SequenceDataFiles(dir=custom_path, child_dirs=[], sequence_name="")
            if custom_path
            else self.data_files[-1]
        )
        if len(custom_path.child_dirs) == 0:
            self.data_files[-1] = self.fuzzy_match_most_recent_folder(
                custom_path
            ).ok_value
        all_w_spectra: List[Dict[int, AgilentHPLCChromatogram]] = []
        for row in self.data_files[-1].child_dirs:
            self.get_uv_spectrum(row)
            all_w_spectra.append(self.uv)
        return all_w_spectra

    def get_data(
        self, custom_path: Optional[str] = None
    ) -> List[AgilentChannelChromatogramData]:
        custom_path = (
            SequenceDataFiles(dir=custom_path, child_dirs=[], sequence_name="")
            if custom_path
            else self.data_files[-1]
        )
        if len(custom_path.child_dirs) == 0:
            self.data_files[-1] = self.fuzzy_match_most_recent_folder(
                custom_path
            ).ok_value
        spectra: List[AgilentChannelChromatogramData] = []
        for row in self.data_files[-1].child_dirs:
            self.get_spectrum_at_channels(row)
            spectra.append(AgilentChannelChromatogramData(**self.spectra))
        return spectra

    def get_report(
        self,
        custom_path: Optional[str] = None,
        report_type: ReportType = ReportType.TXT,
    ) -> List[AgilentReport]:
        if custom_path:
            self.data_files.append(
                self.fuzzy_match_most_recent_folder(
                    most_recent_folder=SequenceDataFiles(
                        dir=custom_path, child_dirs=[], sequence_name="NA"
                    )
                ).ok_value
            )
        parent_dir = self.data_files[-1]
        spectra = self.get_data()
        reports = []
        for i, child_dir in enumerate(parent_dir.child_dirs):
            metd_report = self.get_report_details(child_dir, report_type)
            child_spectra: List[AgilentHPLCChromatogram] = list(
                spectra[i].__dict__.values()
            )
            for j, signal in enumerate(metd_report.signals):
                possible_data = child_spectra[j]
                if len(possible_data.x) > 0:
                    signal.data = possible_data
            reports.append(metd_report)
        return reports



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Union


class Num(Enum):
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8
    NINE = 9

    @classmethod
    def from_num(cls, num: int):
        num_mapping = {
            1: Num.ONE,
            2: Num.TWO,
            3: Num.THREE,
            4: Num.FOUR,
            5: Num.FIVE,
            6: Num.SIX,
            7: Num.SEVEN,
            8: Num.EIGHT,
            9: Num.NINE,
        }

        if num in num_mapping:
            return num_mapping[num]
        else:
            raise ValueError("Num must be between 1 and 9")


class Plate(Enum):
    ONE = -96
    TWO = 4000

    @classmethod
    def from_num(cls, plate: int) -> Plate:
        if 1 <= plate <= 2:
            return Plate.ONE if plate == 1 else Plate.TWO
        raise ValueError("Plate is one or 1 or 2")


class Letter(Enum):
    A = 4191
    B = 4255
    C = 4319
    D = 4383
    E = 4447
    F = 4511

    @classmethod
    def from_str(cls, let: str) -> Letter:
        letter_mapping = {
            "A": Letter.A,
            "B": Letter.B,
            "C": Letter.C,
            "D": Letter.D,
            "E": Letter.E,
            "F": Letter.F,
        }

        if let in letter_mapping:
            return letter_mapping[let]
        else:
            raise ValueError("Letter must be one of A to F")


@dataclass
class FiftyFourVialPlate:
    """
    Class to represent the 54 vial tray. Assumes you have:
    2 plates (P1 or P2)
    6 rows (A B C D E F)
    9 columns (1 2 3 4 5 6 7 8 9)

    valid vial locations: P1-A2, P2-F9
    invalid vial locations: P3-A1, P1-Z3, P2-B10
    """

    plate: Plate
    letter: Letter
    num: Num

    def value(self) -> int:
        return self.plate.value + self.letter.value + self.num.value

    @classmethod
    def from_str(cls, loc: str):
        """
        Converts a string representing the vial location into numerical representation for Chemstation.

        :param loc: vial location
        :return: `FiftyFourVialPlate` object representing the vial location
        :raises: ValueError if string is invalid tray location
        """
        if len(loc) != 5:
            raise ValueError(
                "Plate locations must be PX-LY, where X is either 1 or 2 and Y is 1 to 9"
            )
        try:
            plate = int(loc[1])
            letter = loc[3]
            num = int(loc[4])
            return FiftyFourVialPlate(
                plate=Plate.from_num(plate),
                letter=Letter.from_str(letter),
                num=Num.from_num(num),
            )
        except Exception:
            raise ValueError(
                "Plate locations must be PX-LY, where X is either 1 or 2 and Y is 1 to 9"
            )

    @classmethod
    def from_int(cls, num: int) -> Tray:
        """
        Converts an integer representation of a vial location to a `FiftyFourVialPlate` or `TenVialColumn` object

        :param num: numerical representation of a vial location
        :return: the proper vial location object
        :raises: ValueError no matching can be made
        """
        if num in range(1, 11):
            return TenVialColumn(num)

        row_starts = [
            # plate 1
            FiftyFourVialPlate.from_str("P1-F1"),
            FiftyFourVialPlate.from_str("P1-E1"),
            FiftyFourVialPlate.from_str("P1-D1"),
            FiftyFourVialPlate.from_str("P1-C1"),
            FiftyFourVialPlate.from_str("P1-B1"),
            FiftyFourVialPlate.from_str("P1-A1"),
            # plate 2
            FiftyFourVialPlate.from_str("P2-F1"),
            FiftyFourVialPlate.from_str("P2-E1"),
            FiftyFourVialPlate.from_str("P2-D1"),
            FiftyFourVialPlate.from_str("P2-C1"),
            FiftyFourVialPlate.from_str("P2-B1"),
            FiftyFourVialPlate.from_str("P2-A1"),
        ]

        # find which row
        possible_row = None
        for i in range(0, 6):
            p1_val = row_starts[i].value()
            p2_val = row_starts[6 + i].value()
            if num >= p2_val:
                possible_row = row_starts[6 + i]
            elif p1_val <= num < row_starts[-1].value():
                possible_row = row_starts[i]
            if possible_row:
                break

        # determine which num
        if possible_row:
            starting_loc = possible_row
            base_val = starting_loc.plate.value + starting_loc.letter.value
            for i in range(1, 10):
                if num - i == base_val:
                    return FiftyFourVialPlate(
                        plate=starting_loc.plate,
                        letter=starting_loc.letter,
                        num=Num.from_num(i),
                    )
        raise ValueError("Number didn't match any location. " + str(num))


class TenVialColumn(Enum):
    """
    Class to represent the 10 vial locations.
    """

    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8
    NINE = 9
    TEN = 10


Tray = Union[FiftyFourVialPlate, TenVialColumn]



import abc
from typing import List, Union, Dict, Optional

from result import Result

from ....analysis.process_report import AgilentReport, ReportType
from ....control.controllers import CommunicationController
from ....control.controllers.tables.table import TableController
from pychemstation.analysis.chromatogram import (
    AgilentChannelChromatogramData,
    AgilentHPLCChromatogram,
)
from ....utils.table_types import T, Table


class DeviceController(TableController, abc.ABC):
    def __init__(
        self, controller: CommunicationController, table: Table, offline: bool
    ):
        super().__init__(
            controller=controller, src=None, data_dirs=[], table=table, offline=offline
        )

    @abc.abstractmethod
    def get_row(self, row: int):
        pass

    def retrieve_recent_data_files(self):
        raise NotImplementedError

    def fuzzy_match_most_recent_folder(self, most_recent_folder: T) -> Result[T, str]:
        raise NotImplementedError

    def get_report(
        self, report_type: ReportType = ReportType.TXT
    ) -> List[AgilentReport]:
        raise NotImplementedError

    def get_data_uv(
        self,
    ) -> Union[
        List[Dict[str, AgilentHPLCChromatogram]], Dict[str, AgilentHPLCChromatogram]
    ]:
        raise NotImplementedError

    def get_data(
        self, custom_path: Optional[str] = None
    ) -> Union[List[AgilentChannelChromatogramData], AgilentChannelChromatogramData]:
        raise NotImplementedError



import random
import unittest

from pychemstation.utils.macro import Command
from pychemstation.utils.method_types import TimeTableEntry
from tests.constants import (
    clean_up,
    set_up_utils,
    DEFAULT_METHOD,
)


class TestMethod(unittest.TestCase):
    """
    These tests should always work with an online controller.
    """

    def setUp(self):
        self.hplc_controller = set_up_utils(254, offline=False, runs=False)
        self.hplc_controller.send(
            Command.SAVE_METHOD_CMD.value.format(
                commit_msg="method saved by pychemstation"
            )
        )
        self.hplc_controller.switch_method(DEFAULT_METHOD)
        self.old_method = self.hplc_controller.load_method()

    def tearDown(self):
        self.hplc_controller.edit_method(self.old_method)
        clean_up(self.hplc_controller)

    def test_update_method_components(self):
        new_flow = random.randint(1, 10) / 10
        self.hplc_controller.method_controller.edit_flow(new_flow=new_flow)
        new_post_time = random.randint(1, 6)
        self.hplc_controller.method_controller.edit_post_time(
            new_post_time=new_post_time
        )
        new_stop_time = random.randint(1, 20)
        self.hplc_controller.method_controller.edit_stop_time(
            new_stop_time=new_stop_time
        )
        new_om = random.randint(1, 50)
        self.hplc_controller.method_controller.edit_initial_om(new_om=new_om)
        method_details = self.hplc_controller.load_method()
        self.assertEqual(method_details.params.flow, new_flow)
        self.assertEqual(method_details.post_time, new_post_time)
        self.assertEqual(method_details.stop_time, new_stop_time)
        self.assertEqual(method_details.params.organic_modifier, new_om)

    def test_update_only_timetable(self):
        start_time_1 = random.randint(1, 10)
        starting_og = random.randint(1, 50)
        timetable_flow = random.randint(1, 10) / 10
        timetable = [
            TimeTableEntry(
                start_time=start_time_1,
                organic_modifer=starting_og,
                flow=timetable_flow,
            ),
            TimeTableEntry(
                start_time=start_time_1 + 5, organic_modifer=78, flow=1 - timetable_flow
            ),
        ]
        self.hplc_controller.method_controller.edit_method_timetable(timetable)
        method_details = self.hplc_controller.load_method()
        self.assertEqual(method_details.timetable, timetable)


if __name__ == "__main__":
    unittest.main()
